package docker

import (
	"fmt"
	"reflect"
	"testing"
)

func Test_stripTagPrefix(t *testing.T) {
	tests := []struct {
		Before string
		After  string
	}{
		{"refs/tags/1.0.0", "1.0.0"},
		{"refs/tags/v1.0.0", "1.0.0"},
		{"v1.0.0", "1.0.0"},
	}

	for _, test := range tests {
		got, want := stripTagPrefix(test.Before), test.After
		if got != want {
			t.Errorf("Got tag %s, want %s", got, want)
		}
	}
}

var allTags = []string{
	"v0.9.0",
	"v1.0.0",
	"v1.0.0-alpha.1",
	"v1.0.0-rc1",
}

func TestDefaultTags(t *testing.T) {
	tests := []struct {
		Before        string
		After         []string
		ExpectedError error
	}{
		{"", nil, ErrRefTypeUnsupported{}},
		{"refs/heads/main", []string{"dev"}, nil},
		{"refs/heads/release/v1.13", []string{"1.13-dev"}, nil},
		{"refs/tags/0.9.0", []string{"0.9", "0.9.0"}, nil},
		{"refs/tags/1.0.0", []string{"1.0", "1.0.0", "1", "latest"}, nil},
		{"refs/tags/v1.0.0", []string{"1.0", "1.0.0", "1", "latest"}, nil},
		{"refs/tags/v1.0.0-alpha.1", []string{"1.0.0-alpha.1"}, nil},
		{"refs/tags/v1.0.0-rc1", []string{"1.0.0-rc1"}, nil},
	}

	for _, test := range tests {
		t.Run(test.Before, func(t *testing.T) {
			tags, err := DefaultTags(test.Before, allTags)
			if err != test.ExpectedError {
				t.Error(fmt.Errorf("expected %v but %v", test.ExpectedError, err))
				return
			}
			got, want := tags, test.After
			if !reflect.DeepEqual(got, want) {
				t.Errorf("Got tag %#v, want %#v", got, want)
			}
		})
	}
}

func TestDefaultTagsError(t *testing.T) {
	tests := []string{
		"refs/tags/x1.0.0",
		"refs/tags/20190203",
	}

	for _, test := range tests {
		_, err := DefaultTags(test, allTags)
		if err == nil {
			t.Errorf("Expect tag error for %s", test)
		}
	}
}

func TestDefaultTagSuffix(t *testing.T) {
	tests := []struct {
		Before        string
		Suffix        string
		After         []string
		ExpectedError error
	}{
		// without suffix
		{
			Before:        "",
			After:         nil,
			ExpectedError: ErrRefTypeUnsupported{},
		},
		{
			Before: "refs/tags/v1.0.0",
			After: []string{
				"1.0",
				"1.0.0",
				"1",
				"latest",
			},
		},
		// with suffix
		{
			Suffix: "linux-amd64",
			After:  []string{"linux-amd64"},
		},
		{
			Before: "refs/tags/v1.0.0",
			Suffix: "linux-amd64",
			After: []string{
				"1.0-linux-amd64",
				"1.0.0-linux-amd64",
				"1-linux-amd64",
				"linux-amd64",
			},
		},
		{
			Suffix: "nanoserver",
			After:  []string{"nanoserver"},
		},
		{
			Before: "refs/tags/v1.9.2",
			Suffix: "nanoserver",
			After: []string{
				"1.9-nanoserver",
				"1.9.2-nanoserver",
			},
		},
		{
			Before: "refs/tags/v18.06.0",
			Suffix: "nanoserver",
			After: []string{
				"18.06-nanoserver",
				"18.06.0-nanoserver",
			},
		},
	}

	for _, test := range tests {
		t.Run(test.Before, func(t *testing.T) {
			tag, err := DefaultTagSuffix(test.Before, test.Suffix, allTags)
			if err != test.ExpectedError {
				t.Error(fmt.Errorf("expected %v but %v", test.ExpectedError, err))
				return
			}
			got, want := tag, test.After
			if !reflect.DeepEqual(got, want) {
				t.Errorf("Got tag %#v, want %#v", got, want)
			}
		})
	}
}

func Test_stripHeadPrefix(t *testing.T) {
	type args struct {
		ref string
	}
	tests := []struct {
		args args
		want string
	}{
		{
			args: args{
				ref: "refs/heads/master",
			},
			want: "master",
		},
	}
	for _, tt := range tests {
		if got := stripHeadPrefix(tt.args.ref); got != tt.want {
			t.Errorf("stripHeadPrefix() = %v, want %v", got, tt.want)
		}
	}
}

func TestUseDefaultTag(t *testing.T) {
	type args struct {
		ref           string
		defaultBranch string
	}
	tests := []struct {
		name string
		args args
		want bool
	}{
		{
			name: "latest tag for default branch",
			args: args{
				ref:           "refs/heads/master",
				defaultBranch: "master",
			},
			want: true,
		},
		{
			name: "build from tags",
			args: args{
				ref:           "refs/tags/v1.0.0",
				defaultBranch: "master",
			},
			want: true,
		},
		{
			name: "skip build for not default branch",
			args: args{
				ref:           "refs/heads/develop",
				defaultBranch: "master",
			},
			want: false,
		},
	}
	for _, tt := range tests {
		if got := UseDefaultTag(tt.args.ref, tt.args.defaultBranch); got != tt.want {
			t.Errorf("%q. UseDefaultTag() = %v, want %v", tt.name, got, tt.want)
		}
	}
}
